cdef enum ErrorCode:
    NO_ERROR = 0
    MAGIC_NUMBER_DOES_NOT_MATCH = 1
    VERSION_NUMBER_DOES_NOT_MATCH = 2
    INITIAL_ERROR_CODE = 3
    ONLY_ONE_OUTCOME_PER_EVENT = 7


ERROR_CODES = """
    NO_ERROR = 0
    MAGIC_NUMBER_DOES_NOT_MATCH = 1
    VERSION_NUMBER_DOES_NOT_MATCH = 2
    INITIAL_ERROR_CODE = 3
    ONLY_ONE_OUTCOME_PER_EVENT = 7
    """
